/*
	Given a start and end integers, and a formatted execution string,
	this will delegate the partitions of the input range to slave
	PEs. Intended for simple frame parallel rendering, but general
	enough for any executable that uses an integer range as input.

	--------------------------------------------------------------

	Copyright (C) 2007, Remik Ziemlinski @ noaa gov
	
	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 2, or (at your option)
	any later version.
	
	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.
	
	You should have received a copy of the GNU General Public License
	along with this program; see the file COPYING.
	If not, write to the Free Software Foundation,
	59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.	

	Bugs:
	- Make string substituion more robust.
	- N/A

	Features Todo:
	- N/A

	20070831 rsz Created.
	20071005 rsz 0.1.0 Added input file with integers feature. Slaves can also access CLI options.
 */

#include <stdlib.h>
#include <stdio.h>
#include "mpi.h"
#include "getopt.h"
#include "string.h"
#include <time.h>

#define UNASSIGNED -4
#define DONE -3
#define NOTASK -1
#define NOTREADY -5
#define NEWTASK 2
#define TAG_RUNCONTROL 3
#define TAG_TASKSEND 4
#define TAG_TASKSTATUS 5

static MPI_Request* taskrequests    = NULL;
static int *petasks = NULL;  /* [PErank] -> taskid */
static int *tasks = NULL;    /* [start1,end1 ... startN,endN] */
static int *taskstatus = NULL; /* >=0 PE | UNASSIGNED | DONE */
static time_t *taskt = NULL; /* Array of start/end times per task. */ 
static int npe = 0;
static int ntasks = 0;
static int rank;
static time_t currentt;
static time_t elapsedt;
static time_t startt; /* When this program began. */

#define USAGE "\
mpir --	Given start and end integers, and a formatted execution string, this will delegate the partitions of the input range to slave PEs. Intended for simple frame parallel rendering, but general enough for any executable that uses an integer range as input. An input file can also be used to process specific integers when sequences don't apply.\n\
\n\
USAGE: mpir [OPTION...] command\n\
\n\
  command            Executable string formatted with keywords:\n\
                     %%start, %%end\n\
  -e, --end N        Range end for processing, i.e. last frame number.\n\
  -f, --format F     Output integer format for start/end, such as %04d.\n\
  -g, --granularity G   Number of tasks for PE (default 10).\n\
  -h, --help         Give this usage message.\n\
      --usage                \n\
  -i, --input FILE   File of integers for random processing.\n\
                     Useful when some sequences have holes.\n\
                     Forces granularity to 1.\n\
  -q, --quiet        Don't print status.\n\
  -s, --start N      Starting number for processing range, i.e. first frame.\n\
  -S, --statusdelay S   Status delay in seconds S.\n\
  -v, --version      Print program version.\n\
  -V, --verbose      Progressively output messages to stdout.\n\
\n\
EXAMPLES: mpirun -np 2 mpir -s 101 -e 1000 'blender -b in.blend -s %%start -e %%end -a'\n\
          echo 1 5 8 > frames.txt; mpirun -np 2 mpir -i frames.txt -g 1 'blender -b in.blend -s %%start -e %%end -a'\n\
\n\
Report bugs to Remik Ziemlinski @ noaa gov.\n\
"

typedef struct MPIROPTS {
	char*   command;  /* (allocated). */
	int     end;
	char*    format;
	int     granularity;
	int     help;           /* give usage insructions */
	char*   input;          /* input filename. */
	int     quiet;
	int     start;          /* Start filename number suffix from this. */
	int     version;        /* give program version */
	int     verbose;   /* Verbose echos to stdout. */
	int     statusdelay; /* Seconds between status. */
} mpiropts;

static struct option const long_options[] =
{ 
	{"end" ,        required_argument,      0, 'e'},
	{"format",      required_argument,      0, 'f'},
	{"granularity" ,required_argument,      0, 'g'},
	{"input",       required_argument,      0, 'i'},
	{"start" ,      required_argument,      0, 's'},
	{"statusdelay" ,required_argument,      0, 'S'},
  {"quiet",       no_argument,            0, 'q'},
	{"verbose",     no_argument,            0, 'V'},
	{"help",        no_argument,            0, 'h'},
	{"usage",       no_argument,            0, 'h'},
	{"version",     no_argument,            0, 'v'},
	{0, 0, 0, 0}
}; 

void initmpiropts(mpiropts* popts)
{
	if (popts == NULL) 
		return;

	popts->command = NULL;
	popts->format = NULL;
	popts->granularity = 10;
	popts->help = 0;
	popts->input = NULL;
	popts->quiet = 0;
	popts->start = 0;
	popts->statusdelay = 10;
	popts->end = 0;
	popts->verbose = 0;
	popts->version = 0;
}
void freempiropts(mpiropts* popts)
{
	if (popts == NULL) return;

	if (popts->command != NULL)
		{
			free(popts->command);
			popts->command = NULL;	
		}

	if (popts->format != NULL)
		{
			free(popts->format);
			popts->format = NULL;	
		}

	if (popts->input != NULL)
		{
			free(popts->input);
			popts->input = NULL;	
		}
}
void printusage()
{
	fprintf(stderr, USAGE);
}
void printversion()
{
	fprintf(stderr, "mpir 0.1.0\n");
	fprintf(stderr, "Copyright (C) 2007 Remik Ziemlinski\n\
\n\
This program comes with NO WARRANTY, to the extent permitted by law.\n\
You may redistribute copies of this program\n\
under the terms of the GNU General Public License.\n"); 
}
int getmpiropts(int argc, char** argv, mpiropts* popts)
{
	int c;       
  
	if (popts == NULL) return -1;
  
	while ( (c = getopt_long(argc, argv, "e:f:g:i:s:S:hvV", long_options, 0))
					!= -1
					)
		switch (c)
			{
			case 'c':
				popts->command = (char*)malloc(sizeof(char)*(strlen(optarg) + 1));
				strcpy(popts->command, optarg);
				break;

			case 'e':
				popts->end = atoi(optarg);
				break;

			case 'f':
				popts->format = (char*)malloc(sizeof(char)*(strlen(optarg) + 1));
				strcpy(popts->format, optarg);
				break;

			case 'g':
				popts->granularity = atoi(optarg);
				break;

			case 'i':
				popts->input = (char*)malloc(sizeof(char)*1024);
				memset(popts->input, '\0', 1024);
				strcpy(popts->input, optarg);
				break;

			case 'q':
				popts->quiet = 1;
				break;

			case 's':
				popts->start = atoi(optarg);
				break;

			case 'S':
				popts->statusdelay = atoi(optarg);
				break;

			case 'v':
				popts->version = 1;
				break;

			case 'V':
				popts->verbose = 1;
				break;
			
			case ':':
				fprintf(stderr, "Error, -%c without argument.\n\n", optopt); 
				popts->help = 1;
				break;
			case '?':
				fprintf(stderr, "Error.  Unknown argument %c.\n\n", optopt);
				popts->help = 1;
				break;
			case 'h':
				popts->help = 1;
				break;
			}

	if (popts->help) {
		printusage();
		return -1;
	}       
	if (popts->version)		{
		printversion();
		return -1;
	} 

	if (popts->format == NULL) {
		popts->format = (char*)malloc(sizeof(char)*5);
		strcpy(popts->format, "%04d");
	}

	argc -= optind;
	argv += optind;
	if ( (argc < 1) ||
			 (argv[0] == NULL) ) {
		fprintf(stderr, "Error. command not given.\n");
		printusage();
		return -1;
	} else {
		popts->command = (char*)malloc(sizeof(char)*1024);
		popts->command[0] = '\0';
		/* Build up the command. */
		for(c=0; c < argc; ++c) {			
			strcat(popts->command, argv[c]);
		}
		#ifdef DEBUG
		printf("popts->command = %s\n", popts->command);
    #endif
	}

	return 0;
}
char* getmacrostr(int i)
{
	switch(i) {
	case UNASSIGNED: return "UNASSIGNED";
	case DONE: return "DONE";
  case NOTASK: return "NOTASK";
  case NOTREADY: return "NOTREADY";
  case NEWTASK: return "NEWTASK";
  case TAG_RUNCONTROL: return "";
	case TAG_TASKSEND: return "";
	case TAG_TASKSTATUS: return "";
	default: return "?";
	}
}
/* str should be pre-allocated. */
void seconds2hms(int sec, char *str)
{
	sprintf(str, "%02d:%02d:%02d", sec/3600, (sec % 3600)/60, ((sec % 3600) % 60));
}
void printtasks(mpiropts *opt)
{
	int i;
	int taskpe[1024];
	char hms[12], minstr[12], maxstr[12], avgstr[12], elapsedstr[12], totalstr[12];	
	double totalt, nframes, ndone, t, mint, maxt;
	time_t curt;

	#ifdef DEBUG
	printf("printtasks\n");
  #endif

	nframes = (opt->input == NULL) ? opt->end - opt->start + 1 : ntasks;

	for(i=1; i<npe; ++i) {
		if ( petasks[i] >= 0) {
			taskpe[ petasks[i] ] = i; 
		}
	}

	printf(  "-----------------------------------------------\n");
	printf(  "Task |  PE  | Start | End   | Time  | Time Avg \n");

	totalt = 0;
	ndone = 0;
	mint = 99999;
	maxt = -1;
	for(i=0; i < ntasks; ++i) {
		switch(taskstatus[i]) {
		case UNASSIGNED:
			printf("%03d  |      | %04d  | %04d  |       |\n",
						 i, tasks[2*i], tasks[2*i+1]);
		break;
		case DONE:
			t = (taskt[2*i+1]-taskt[2*i]);

			printf("%03d  | DONE | %04d  | %04d  | %02d:%02d | %02d:%02d\n",
						 i, tasks[2*i], tasks[2*i+1],
						 (int)(t/60), (int)t % 60, 
						 (int)(t/(tasks[2*i+1]-tasks[2*i]+1) / 60.),
						 (int)(t/(tasks[2*i+1]-tasks[2*i]+1)) % 60);
			totalt += t;
			ndone += tasks[2*i+1]-tasks[2*i]+1;
			if (mint > (t/(tasks[2*i+1]-tasks[2*i]+1))) mint = t/(tasks[2*i+1]-tasks[2*i]+1);
			if (maxt < (t/(tasks[2*i+1]-tasks[2*i]+1))) maxt = t/(tasks[2*i+1]-tasks[2*i]+1);
			break;
		default: /* Executing. */
			printf("%03d  | %04d | %04d  | %04d  |       |\n",
						 i, taskpe[i], tasks[2*i], tasks[2*i+1]);
			break;
		}
	}

	printf(  "------------------------------------------------------\n");
	printf(  "Elapsed  | Computed |   Min    |   Avg    |   Max    | Progress\n");

	if (mint == 99999) mint = 0;
	if (maxt < 0) maxt = 0;

	time(&curt);
	seconds2hms(curt-startt, elapsedstr);
	seconds2hms(totalt, totalstr);
	seconds2hms(mint, minstr);
	seconds2hms(totalt/(ndone==0?1:ndone), avgstr);
	seconds2hms(maxt, maxstr);
	printf("%s | %s | %s | %s | %s | %d/%d (%d%%)\n", elapsedstr, 
				 totalstr, minstr, avgstr, maxstr, 
				 (int)ndone, (int)(nframes), 
				 (int)((100*ndone)/(double)nframes)
				 );
}
/* Returns task id, or -1 if none left. */
int getavailtask()
{
    int i;
    
    for(i = 0; i < ntasks; ++i) {
			/* is it done or pending? */
			if (taskstatus[i] == UNASSIGNED) 
				return i;
    }
    
    return -1;
}
void sendtask(int tid, int pe)
{
	int buf = NEWTASK;

	#ifdef DEBUG
	printf("sendtask(tid=%d, pe=%d)\n", tid, pe);
  #endif

	if (tid < 0) return;

  #ifdef DEBUG
	printf("Master -> %d: send NEWTASK\n", pe);
  #endif	

	MPI_Isend(  &buf, 1, 
							MPI_INT, pe, TAG_RUNCONTROL, 
						 MPI_COMM_WORLD, &taskrequests[pe]);
	MPI_Request_free(&taskrequests[pe]);
  #ifdef DEBUG
	printf("Master -> %d: send NEWTASK Done\n", pe);
  #endif

  #ifdef DEBUG
	printf("Master -> %d: send start=%d\n", pe, tasks[2*tid]);
  #endif
	MPI_Isend(  &tasks[2*tid], 1, 
							MPI_INT, pe, TAG_TASKSEND, 
						 MPI_COMM_WORLD, &taskrequests[pe]);
	MPI_Request_free(&taskrequests[pe]);
  #ifdef DEBUG
	printf("Master -> %d: send start=%d done\n", pe, tasks[2*tid]);
  #endif
	MPI_Isend(  &tasks[2*tid+1], 1, 
							MPI_INT, pe, TAG_TASKSEND, 
							MPI_COMM_WORLD, &taskrequests[pe]);
	MPI_Request_free(&taskrequests[pe]);	
  #ifdef DEBUG
	printf("Master -> %d: send end=%d done\n", pe, tasks[2*tid+1]);
  #endif

	/* Record start time. */
	taskt[tid*2] = time(NULL);
}
int assigntasks()
{
	int pe, tid;

	#ifdef DEBUG
	printf("assigntasks\n");
  #endif

	for(pe = 1; pe < npe; ++pe)  {
		if(petasks[pe] == NOTREADY) 
			continue;

		if(petasks[pe] < 0) {
			/* assign new task */
			tid = getavailtask();
			
			#ifdef DEBUG
			printf("Master: Next available task id = %d\n", tid);
			#endif

			if (tid < 0)
				return 0; /* No more tasks. */

			petasks[pe] = tid; 
			sendtask(tid, pe);
			taskstatus[tid] = pe;

			#ifdef DEBUG
			printf("\tpetasks[%d]=%d; taskstatus[%d]=%d\n",
						 pe, tid, tid, pe);						 
			#endif
		}
	}
  #ifdef DEBUG
	printf("assigntasks exit\n");
	#endif

	return 0;
}
void shutdownslaves()
{
	int i;
	int buf = DONE;

	#ifdef DEBUG
	printf("Shutting down...\n");
  #endif

	for(i=1; i < npe; ++i) {
		MPI_Isend(  &buf, 1, 
								MPI_INT, i, TAG_RUNCONTROL, 
								MPI_COMM_WORLD, &taskrequests[i]);
		MPI_Request_free(&taskrequests[i]);	
	}
}
void tryshutdownslaves()
{
	int i;

  #ifdef DEBUG
	printf("tryshutdownslaves\n");
  #endif

	/* If all tasks done, send terminate signal. */
	for(i=0; i < ntasks; ++i) {
		if (taskstatus[i] != DONE) {
			#ifdef DEBUG
			printf("tryshutdownslaves abort\n");
      #endif

			return;
		}
	}

	shutdownslaves();

	#ifdef DEBUG
	printf("tryshutdownslaves exit\n");
  #endif
}
void checkandassigntasks(mpiropts *opt)
{
	int status, pe;
	MPI_Status mpistatus;
	/* Allow for ready signals from slaves, so we can use the same loop
	   to prime the work.
	*/
	int ndone = -npe+1;  

  #ifdef DEBUG
	printf("checkandassigntasks\n");
  #endif

	do {
		/* Find tasks done and then assign idle PE new task. */
		for(pe = 1; pe < npe; ++pe) {
			#ifdef DEBUG
      printf("Iprobe\n");
      #endif
			/* message waiting available for master? */
			MPI_Iprobe(pe, TAG_TASKSTATUS, MPI_COMM_WORLD, 
								 &status, &mpistatus);
			#ifdef DEBUG
      printf("Iprobe status=%s\n", (status==1)?"Available":"N/A");
			#endif

			if(status != 1) continue; /* Check other PEs. */

			#ifdef DEBUG
			printf("MASTER Recv: status\n");
			#endif
			
			/* receive task status */
			MPI_Recv((void*)&status, 1, MPI_INT, pe, 
							 TAG_TASKSTATUS, MPI_COMM_WORLD, &mpistatus);
      #ifdef DEBUG
			printf("MASTER Recv: status=%s done\n", getmacrostr(status));
      #endif

			if(status == DONE) {
					/* Index of task (its ID). 
						 Less than 0 means no task was available. */
				if (petasks[pe] >= 0) {
					#ifdef DEBUG
					printf("\tpetasks[%d]==%d\n",pe,petasks[pe]);
					#endif
					taskstatus[ petasks[pe] ] = DONE;
					
					/* Record end time. */
					taskt[petasks[pe]*2+1] = time(NULL);
				}
				/* Increment in all cases to handle spin-up. */
				++ndone; 
				/* Mark PE as ready for a new task. */
				petasks[pe] = NOTASK; 
			}
		}

		assigntasks();
		sleep(1);

		if (!opt->quiet) {
			time(&currentt);
			if (difftime(currentt, elapsedt) > opt->statusdelay) {
				printtasks(opt);
				elapsedt = currentt;
			}
		}
	} while(ndone != ntasks);

	if (opt->verbose) {
		printtasks(opt);
	}		

	tryshutdownslaves();
}
void awaittask(mpiropts *opt)
{
	int start, end, cmdlen, status, len;
	MPI_Status mpistatus;
  char cmd[1024], cmdtemplate[1024], chartmp[16];
	char *replace, *prevreplace;
	int offset = 0, i;

	MPI_Recv(&status, 1, MPI_INT, 0, TAG_RUNCONTROL, 
					 MPI_COMM_WORLD, &mpistatus);
	/* Problem starting according to master, so abort. */
	if(status == DONE) return;

  #ifdef DEBUG
	printf("%d Recv: status=%s\n",rank, getmacrostr(status));
  #endif

	MPI_Recv(&cmdlen, 1, MPI_INT, 0, TAG_TASKSEND, 
					 MPI_COMM_WORLD, &mpistatus);

  #ifdef DEBUG
	printf("%d Recv: cmdlen=%d\n",rank,cmdlen);
  #endif

	MPI_Recv(&cmdtemplate, cmdlen, MPI_CHAR, 0, TAG_TASKSEND, 
					 MPI_COMM_WORLD, &mpistatus);
  #ifdef DEBUG
	printf("%d Recv: cmdtemplate=%s\n",rank,cmdtemplate);
  #endif

	status = DONE;
	MPI_Send(&status, 1, MPI_INT, 0, 
					 TAG_TASKSTATUS, MPI_COMM_WORLD);

  #ifdef DEBUG
	printf("%d Entering loop\n",rank); 
  #endif
	while(1) {
		MPI_Recv(&status, 1, MPI_INT, 0, TAG_RUNCONTROL, 
						 MPI_COMM_WORLD, &mpistatus);
    #ifdef DEBUG
  	printf("%d Recv: status=%s\n",rank, getmacrostr(status));
    #endif
	
		if(status == DONE) break;

		if(status == NEWTASK) {
			memset(cmd,'\0',1024);
			memset(chartmp,'\0',16);

			MPI_Recv(&start, 1, MPI_INT, 0, TAG_TASKSEND, 
							 MPI_COMM_WORLD, &mpistatus);
      #ifdef DEBUG
     	printf("%d Recv: start=%d\n",rank,start);
      #endif
			MPI_Recv(&end, 1, MPI_INT, 0, TAG_TASKSEND, 
							 MPI_COMM_WORLD, &mpistatus);
      #ifdef DEBUG
     	printf("%d Recv: end=%d\n",rank,end);
      #endif

			prevreplace = cmdtemplate;
			replace = strstr(cmdtemplate, "%start");
			if (replace != NULL) {
				offset = replace - cmdtemplate;
				/* Copy text up to the %start token. */
				strncat(cmd, cmdtemplate, offset);
				sprintf(chartmp, opt->format, start);
				strcat(cmd, chartmp);
				/* Next position in command template. */
				prevreplace = replace + 6;
			}
			 
			memset(chartmp,'\0',16);
			replace = strstr(prevreplace, "%end");			
			if (replace != NULL) {
				offset = replace - prevreplace;
				/* Append text in command between %start and %end. */
				strncat(cmd, prevreplace, offset);
				sprintf(chartmp, opt->format, end);
				len = strlen(chartmp);
				strcat(cmd, chartmp);
				/* Append remainder of command. */
				strcat(cmd, replace + 4);
			} else {
				/* No occurence of %end, just append remainder. */
				strcat(cmd, prevreplace);
			}

			/* Exec command. */
			if (opt->verbose) {
				printf("%s\n", cmd);
			}

			system(cmd);
			status = DONE;
			MPI_Send(&status, 1, MPI_INT, 0, 
							 TAG_TASKSTATUS, MPI_COMM_WORLD);
		}
	}
}
int scatter(mpiropts *opt)
{
	int i, cmdlen, buf;
	FILE* f;

	cmdlen = strlen(opt->command);

	petasks = (int*)malloc(sizeof(int)*npe);
	
	/* If not using input file, use range from command line. */
	if (opt->input == NULL) {
		ntasks = (opt->end - opt->start + 1)/opt->granularity;
		tasks = (int*)malloc(sizeof(int)*ntasks*2);
	} else {
		/* Read the input file. */
		f = fopen(opt->input, "r");
		if (f == NULL)  {
			printf("Error opening file %s. Aborting.\n", opt->input);
			exit(1);
		}
		
		/* Count number of tasks. */
		ntasks = 0;
		while (EOF != fscanf(f, "%d\n",&i)) ++ntasks;
		fclose(f);

		f = fopen(opt->input, "r");
		tasks = (int*)malloc(sizeof(int)*ntasks*2);
		i = 0;
		while (EOF != fscanf(f, "%d\n", &tasks[2*i])) {
			tasks[2*i+1] = tasks[2*i];
			++i;
		}

		fclose(f);

		/* Force granularity to 1 since each integer in input file will be a task. */
		opt->granularity = 1;
	}

	taskstatus = (int*)malloc(sizeof(int)*ntasks);
	taskrequests = (MPI_Request*)malloc(sizeof(MPI_Request)*npe);
	taskt = (time_t*)malloc(sizeof(time_t)*ntasks*2);

	time(&startt);
	elapsedt = startt;

	if (opt->input == NULL) {
		/* Define task start/end frames from range. */
		for(i=0; i < ntasks-1; ++i) {
			tasks[2*i] = opt->start + i * opt->granularity;
			tasks[2*i+1] = opt->start + (i+1) * opt->granularity - 1;
		}
		tasks[2*(ntasks-1)  ] = opt->start + (ntasks-1) * opt->granularity;
		tasks[2*(ntasks-1)+1] = opt->end;
	} 

	for(i=0; i < ntasks; ++i) {
		taskstatus[i] = UNASSIGNED;
	}

	petasks[0] = NOTASK;
	buf = NEWTASK;
	for(i=1; i < npe; ++i) {
		petasks[i] = NOTREADY; /* No task id assigned to PE. */
		
		#ifdef DEBUG
		printf("Master -> %d: Activate\n",i);
		#endif
		/* Give okay to slave to start. */
		MPI_Send(  &buf, 1, 
								MPI_INT, i, TAG_RUNCONTROL, 
								MPI_COMM_WORLD);

		/* Send command template. */
		#ifdef DEBUG
		printf("Master -> %d: cmdlen=%d\n",i,cmdlen);
		#endif
		MPI_Send(  &cmdlen, 1, 
								MPI_INT, i, TAG_TASKSEND, 
							 MPI_COMM_WORLD);
		MPI_Send(  opt->command, cmdlen, 
								MPI_CHAR, i, TAG_TASKSEND, 
							 MPI_COMM_WORLD);
	}


	checkandassigntasks(opt);

	free(petasks);
	free(tasks);
	free(taskstatus);
	free(taskrequests);
	free(taskt);
	
	return 0;
}
int main(int argc, char** argv)
{
	int status;
	mpiropts opts;
	MPI_Init(&argc,&argv);
	MPI_Comm_rank(MPI_COMM_WORLD, &rank);
	MPI_Comm_size(MPI_COMM_WORLD, &npe);

	initmpiropts(&opts);
	status = getmpiropts(argc, argv, &opts);

	if (rank == 0) {	
		if (status < 0)
			shutdownslaves();
		else
			status = scatter(&opts);               
		
	} else {
		awaittask(&opts);
	}

	freempiropts(&opts); 

	MPI_Finalize();
	
	return status;
}
